import 'package:bizdoc/gen_l10n/app_localizations.dart';
import 'package:flutter/material.dart';
import 'package:pull_to_refresh/pull_to_refresh.dart';
import 'package:timeago/timeago.dart' as time_ago;

import 'package:bizdoc/core/hub.dart';
import 'package:bizdoc/core/extensions.dart';
import 'package:bizdoc/models/profile.dart';
import 'package:bizdoc/services/mailbox.dart';
import 'package:bizdoc/services/session.dart';
import 'package:bizdoc/models/contracts.dart';
import 'package:bizdoc/models/data.dart';
import 'compose.dart';

/// List [HeaderModel] messages
class BrowsePage extends StatefulWidget {
  static const routeName = 'mailbox/f';
  final String? folderId;
  final QueryType? queryType;
  final Function(Iterable<ActionInfo>)? onSelected;
  BrowsePage({this.folderId, this.queryType, this.onSelected})
      : super(
          key: Key(folderId ?? queryType.toString()),
        );
  @override
  _BrowsePageState createState() => _BrowsePageState();
}

class _BrowsePageState extends State<BrowsePage> {
  final _items = <HeaderModel>[];
  final _selectedItems = <RecipientModel>[];
  final _actions = <String>{};

  final RefreshController _refreshController =
      RefreshController(initialRefresh: false);

  //final oCcy = new NumberFormat("#,##0.00", "en_US");
  @override
  void initState() {
    _listen();
    _onLoading();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return SmartRefresher(
      enablePullDown: true,
      enablePullUp: true,
      header: WaterDropHeader(
        complete: Container(),
      ),
      footer: CustomFooter(
        builder: (BuildContext context, LoadStatus? mode) {
          Widget body;
          if (mode == LoadStatus.idle) {
            body = Container();
          } else if (mode == LoadStatus.loading) {
            body = const CircularProgressIndicator();
          } else if (mode == LoadStatus.failed) {
            body = Container();
          } else if (mode == LoadStatus.canLoading) {
            body = Text(AppLocalizations.of(context)!.loadMore);
          } else {
            body = Container();
          }
          return SizedBox(
            height: 55.0,
            child: Center(child: body),
          );
        },
      ),
      controller: _refreshController,
      onRefresh: _onRefresh,
      onLoading: _onLoading,
      child: ListView.separated(
        itemBuilder: (context, index) {
          final item = _items[index];
          return ListTile(
              onLongPress: () => _long(item),
              key: Key(item.id.toString()),
              title: Text(item.number),
              subtitle: Text(time_ago.format(item.received)),
              onTap: () => _open(item),
              trailing: IconButton(
                icon: item.flag
                    ? const Icon(Icons.flag)
                    : const Icon(Icons.flag_outlined),
                onPressed: () => _toggleFlag(item),
              ));
        },
        separatorBuilder: (context, index) {
          return const Divider();
        }, // itemExtent: 100.0,
        itemCount: _items.length,
      ),
    );
  }

  void _onRefresh() async {
    try {
      final items = await mailbox.messages(
          folderId: widget.folderId, queryType: widget.queryType);
      _items.clear();
      _items.addAll(items);
      _refreshController.refreshCompleted();
    } catch (e) {
      _refreshController.refreshFailed();
      _err();
    }
  }

  void _err() => _snackBar(AppLocalizations.of(context)!.err);

  void _onLoading() async {
    try {
      final items = await mailbox.messages(
          folderId: widget.folderId,
          queryType: widget.queryType,
          skip: _items.length);
      if (items.isNotEmpty) {
        _items.addAll(items);
        _refreshController.loadComplete();
      } else {
        _refreshController.loadNoData();
      }
    } catch (e) {
      _refreshController.loadFailed();
      _err();
    }
    if (mounted) setState(() {});
  }

  Future<void> _long(HeaderModel item) async {
    if (item is RecipientModel) {
      _selectedItems.remove(item);
    } else {
      try {
        final model = await mailbox.get(item.id);
        final index = _items.indexOf(item);
        _items.replaceRange(index, index, [model]);
        _selectedItems.add(model);
      } catch (e) {
        _err();
        return;
      }
    }
    _actions.clear();
    if (_selectedItems.isEmpty) return;
    if (_selectedItems.any(
        (element) => element.actions == null || element.actions!.isEmpty)) {
      return;
    }
    _actions.addAll(_selectedItems.first.actions!);
    for (var i = 0; i < _selectedItems.length; i++) {
      _actions.retainAll(_selectedItems[i].actions!);
    }
    Iterable<ActionInfo> actions;
    if (_actions.length == 1) {
      final action = session.profile.actions
          .firstWhere((element) => element.name == _actions.first);
      actions = [action];
    } else {
      actions = session.profile.actions.where((element) =>
          element.multiple != false && _actions.contains(element.name));

      widget.onSelected!(actions);
    }
  }

  _action(ActionInfo action) async {
    final args = await ActionDialog.show(context, action);
    if (args != null) {
      for (var item in _selectedItems) {
        try {
          await mailbox.send(item.id, item.version, action.name, args);
        } catch (e) {
          _err();
        }
      }
      _snackBar(_selectedItems.length > 1
          ? AppLocalizations.of(context)!.itemsAction.format([
              _selectedItems.length.toString(),
              action.adjective ?? action.past ?? action.title
            ])
          : '${_selectedItems.first.number} ' +
              (action.adjective ?? action.past ?? action.title));
    }
  }

  _open(HeaderModel item) async {
    try {
      if (item is RecipientModel) {
        Navigator.pushNamed(context, ComposePage.routeName,
            arguments: ComposeModel(item));
      } else {
        final model = await mailbox.get(item.id);
        Navigator.pushNamed(context, ComposePage.routeName,
            arguments: ComposeModel(model));
      }
    } catch (e) {
      _err();
    }
  }

  void _toggleFlag(HeaderModel item) async {
    try {
      await mailbox.flag(item.id, !item.flag);
      setState(() {
        item.flag = !item.flag;
      });
    } catch (e) {
      _err();
    }
  }

  void _snackBar(String message) => ScaffoldMessenger.of(context)
      .showSnackBar(SnackBar(content: Text(message)));

  Future<void> _listen() async {
    await for (var e in hub.events) {
      if (e is NewMail && e.model.folderId == widget.folderId) {
        _items.insert(0, e.model);
      } else if (e is MailUpdate) {
        final found =
            _items.firstWhereOrNull((element) => element.id == e.model.id);
        if (found != null) {
          found.update(e.model);
        } else if (e.model.folderId != widget.folderId) {
          _items.remove(found);
        }
      }
    }
  }
}
