import 'package:bizdoc/models/cube.dart';
import 'package:flutter/material.dart';
import 'package:intl/intl.dart';
import 'package:syncfusion_flutter_charts/charts.dart';
import 'package:syncfusion_flutter_datagrid/datagrid.dart';

import 'package:bizdoc/core/extensions.dart';
import 'package:bizdoc/services/cube.dart';
import 'package:bizdoc/services/datasource.dart';
import 'package:bizdoc/models/profile.dart';
import 'explore.dart';

/// show cube view of either [ChartView], [PivotView] or [GridView]. Navigate to [ExplorePage] on tap.
class CubeViewPage extends StatelessWidget {
  final CubeInfo cube;
  final CubeView view;
  final Map<String, dynamic>? filters;
  static const routeName = 'cube';
  const CubeViewPage(
      {Key? key, required this.cube, required this.view, this.filters})
      : super(key: key);
  @override
  Widget build(BuildContext context) {
    switch (view.type) {
      case ViewType.grid:
        return GridView(cube, view, filters: filters);
      case ViewType.pivot:
        return PivotView(cube, view, filters: filters);
      default:
        return ChartView(cube, view, filters: filters);
    }
  }
}

///
class GridView extends StatefulWidget {
  final CubeInfo cube;
  final CubeView view;
  final Map<String, dynamic>? filters;
  const GridView(this.cube, this.view, {Key? key, this.filters})
      : super(key: key);
  @override
  _GridViewState createState() => _GridViewState();
}

class _GridViewState extends State<GridView> {
  late CubeAxis _x;
  late CubeAxis _y;

  late Map<String, dynamic> _columns;
  late Map<String, dynamic> _rows;

  final _oCcy = NumberFormat("###,###,##0.00");

  int _lines = 0;

  @override
  void initState() {
    _oCcy.currencyName = widget.cube.currencyCode;
    _x = widget.cube.axes
        .firstWhere((element) => element.name == widget.view.xAxis.first);
    _y = widget.cube.axes
        .firstWhere((element) => element.name == widget.view.series!.first);
    super.initState();
  }

  Future<List<Map>> _initializeData() async {
    /// retrieve data
    final data = await cube.series(widget.cube.name, [_x.name],
        series: [_y.name],
        indices: widget.view.indices,
        filters: widget.filters);

    /// initialize columns and rows
    _columns = await dataSource.map(_x.dataType);
    _rows = await dataSource.map(_y.dataType);

    final rows = <Map>[];

    /// index series into first rows
    for (var series in data.where((element) => element.index != null)) {
      final Map row = {'title': series.title, 'name': series.name};
      for (var element in series.points) {
        row[element.x] = element.y;
      }
      _lines++;
      rows.add(row);
    }

    /// series of rows axes
    for (var key in _rows.keys) {
      final series = data.firstWhereOrNull((element) => element.name == key);
      final Map row = {'title': _rows[key], 'name': key};
      if (series != null) {
        for (var element in series.points) {
          row[element.x] = element.y;
        }
      }
      rows.add(row);
    }
    return rows;
  }

  /// build [SfDataGrid]
  @override
  Widget build(BuildContext context) {
    return FutureBuilder<List<Map>>(
      future: _initializeData(),
      builder: (context, snapshot) {
        if (snapshot.hasData) {
          return SfDataGrid(
            frozenColumnsCount: 1,
            frozenRowsCount: _lines,
            onCellTap: (details) {
              final Map<String, dynamic> axes = {};
              axes[_x.name] =
                  _columns.keys.elementAt(details.rowColumnIndex.columnIndex);
              axes[_y.name] =
                  _columns.keys.elementAt(details.rowColumnIndex.rowIndex);
              _nav(axes, null);
            },
            columns: [
              GridColumn(columnName: 'title', label: Text(_y.title ?? '')),
              for (var column in _columns.entries)
                GridColumn(columnName: column.key, label: Text(column.value)),
            ],
            source: CubeDataSource(snapshot.data!),
          );
          // return Table(
          //   children: [
          //     for (var item in snapshot.data!)
          //       TableRow(
          //         children: [
          //           TableCell(child: Text(item['title'])),
          //           for (var key in _columns.keys)
          //             TableCell(child: Text(item[key]))
          //         ],
          //       )
          //   ],
          // );
        }
        return const CircularProgressIndicator();
      },
    );
  }

  _nav(Map<String, dynamic> axes, String? index) {
    Navigator.pushNamed(context, ExplorePage.routeName,
        arguments: ExploreArgs(widget.cube.name, axes: axes, index: index));
  }
}

class CubeDataSource extends DataGridSource {
  final List<Map> data;
  CubeDataSource(this.data);
  @override
  DataGridRowAdapter? buildRow(DataGridRow row) {
    return DataGridRowAdapter(
        cells: row.getCells().map((e) => Text(e.value.toString())).toList());
  }
}

/// build [SfDataGrid]
class PivotView extends StatefulWidget {
  final CubeInfo cube;
  final CubeView view;
  final Map<String, dynamic>? filters;
  const PivotView(this.cube, this.view, {Key? key, this.filters})
      : super(key: key);
  @override
  _PivotViewState createState() => _PivotViewState();
}

class _PivotViewState extends State<PivotView> {
  late List<String> _axes;
  final _oCcy = NumberFormat("###,###,##0.00");

  @override
  void initState() {
    _oCcy.currencyName = widget.cube.currencyCode;
    _axes = widget.view.xAxis.toList();
    if (widget.view.series != null) _axes.addAll(widget.view.series!);
    super.initState();
  }

  Future<List<Map>> _initializeData() async {
    try {
      final data = await cube.table(widget.cube.name, _axes,
          indices: widget.view.indices, filters: widget.filters);
      return data;
    } catch (e) {
      rethrow;
    }
  }

  @override
  Widget build(BuildContext context) {
    return FutureBuilder<List<Map>>(
      future: _initializeData(),
      builder: (context, snapshot) {
        if (snapshot.hasData) {
          return Table(
            children: [
              for (var item in snapshot.data!)
                TableRow(
                  children: [
                    for (var axis in _axes)
                      TableCell(child: Text(item[axis + 'Label'])),
                    for (var index in widget.view.indices ?? [])
                      TableCell(child: Text(item[index])),
                    TableCell(child: Text(_oCcy.format(item['value'])))
                  ],
                )
            ],
          );
        }
        return const CircularProgressIndicator();
      },
    );
  }

  _nav(Map<String, dynamic> axes) {
    Navigator.pushNamed(context, ExplorePage.routeName,
        arguments: ExploreArgs(widget.cube.name, axes: axes));
  }
}

/// build [SfCircularChart] or [SfCartesianChart]. navigate to [ExplorePage]
class ChartView extends StatefulWidget {
  final CubeInfo cube;
  final CubeView view;
  final Map<String, dynamic>? filters;
  const ChartView(this.cube, this.view, {Key? key, this.filters})
      : super(key: key);
  @override
  _ChartViewState createState() => _ChartViewState();
}

class _ChartViewState extends State<ChartView> {
  late Future<List<CubeSeries>> _data;

  @override
  void initState() {
    _data = cube.series(widget.cube.name, widget.view.xAxis,
        series: widget.view.series,
        indices: widget.view.indices,
        filters: widget.filters);
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return FutureBuilder<List<CubeSeries>>(
      future: _data,
      builder: (context, snapshot) {
        if (snapshot.hasData) {
          switch (widget.view.chartType) {
            case ChartType.pie:
              return SfCircularChart(
                series: [
                  PieSeries<Point, String>(
                    dataSource: snapshot.data![0].points,
                    xValueMapper: (datum, index) => datum.xLabel,
                    yValueMapper: (datum, index) => datum.y,
                  )
                ],
                onSelectionChanged: (pointTapArgs) =>
                    _explore(pointTapArgs, context),
              );
            case ChartType.doughnut:
            case ChartType.halfDoughnut:
              return SfCircularChart(
                series: [
                  DoughnutSeries<Point, String>(
                    startAngle:
                        widget.view.chartType == ChartType.doughnut ? 360 : 270,
                    endAngle:
                        widget.view.chartType == ChartType.doughnut ? 0 : 90,
                    dataSource: snapshot.data![0].points,
                    xValueMapper: (datum, index) => datum.xLabel,
                    yValueMapper: (datum, index) => datum.y,
                  )
                ],
                onSelectionChanged: (pointTapArgs) =>
                    _explore(pointTapArgs, context),
              );
            default:
              final series = <ChartSeries>[];
              for (var data in snapshot.data!) {
                switch (widget.view.chartType) {
                  case ChartType.bar:
                    series.add(BarSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingBar:
                    series.add(StackedBarSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingBar100:
                    series.add(StackedBar100Series<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.area:
                    series.add(AreaSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.spline:
                    series.add(SplineSeries(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.splineArea:
                    series.add(SplineAreaSeries(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stepLine:
                    series.add(StepLineSeries(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stepArea:
                    series.add(StepAreaSeries(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingColumn:
                    series.add(StackedColumnSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingColumn100:
                    series.add(StackedColumn100Series<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingArea:
                    series.add(StackedAreaSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingArea100:
                    series.add(StackedArea100Series<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingLine:
                    series.add(StackedLineSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.stackingLine100:
                    series.add(StackedLine100Series<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.scatter:
                    series.add(ScatterSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.bubble:
                    series.add(BubbleSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.pyramid:
                    series.add(PyramidSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  case ChartType.funnel:
                    series.add(FunnelSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                    break;
                  default:
                    series.add(LineSeries<Point, String>(
                      dataSource: data.points,
                      name: data.title,
                      xValueMapper: (datum, index) => datum.xLabel,
                      yValueMapper: (datum, index) => datum.y,
                    ));
                }
              }
              return SfCartesianChart(
                primaryXAxis: CategoryAxis(),
                primaryYAxis: NumericAxis(),
                legend: Legend(isVisible: true),
                series: series,
                onSelectionChanged: (pointTapArgs) => _explore(
                    pointTapArgs, context,
                    series: snapshot.data!.elementAt(pointTapArgs.seriesIndex)),
              );
          }
        }
        return const LinearProgressIndicator();
      },
    );
  }

  ///
  void _explore(SelectionArgs pointTapArgs, BuildContext context,
      {CubeSeries? series}) {
    final point = series!.points[pointTapArgs.pointIndex];
    final Map<String, dynamic> axes = {};
    for (var i = 0; i < widget.view.xAxis.length; i++) {
      axes[widget.view.xAxis[i]] = point.x.split(complexAxisDivider)[i];
    }
    if (widget.view.series != null && series.name != null) {
      for (var i = 0; i < widget.view.series!.length; i++) {
        axes[widget.view.series![i]] =
            series.name!.split(complexAxisDivider)[i];
      }
    }
    Navigator.pushNamed(context, ExplorePage.routeName,
        arguments: ExploreArgs(widget.cube.name, axes: axes));
  }
}
