import 'package:bizdoc/controls/address.dart';
import 'package:bizdoc/controls/autocomplete.dart';
import 'package:bizdoc/controls/select.dart';
import 'package:bizdoc/models/profile.dart';
import 'package:date_field/date_field.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

/// Layout input from [FieldInfo] list. Trigger [onChanged]
class Fields extends StatefulWidget {
  final List<FieldInfo> fields;
  final Function onChanged;
  const Fields({Key? key, required this.fields, required this.onChanged})
      : super(key: key);

  @override
  _FieldsState createState() => _FieldsState();
}

class _FieldsState extends State<Fields> {
  @override
  Widget build(BuildContext context) {
    final args = {};
    final List<Widget> children = [];
    for (var field in widget.fields) {
      switch (field.kind) {
        case FieldKind.select:
          children.add(Text(field.label ?? field.name));
          children.add(SelectField(
            dataType: field.source,
            onTap: (value) {
              args[field.name] = value;
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.autocomplete:
          children.add(Text(field.label ?? field.name));
          children.add(AutocompleteField(
            dataType: field.source,
            onChanged: (key, value) {
              args[field.name] = key;
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.date:
          final DateTime date = DateTime.now();
          children.add(InputDatePickerFormField(
            firstDate: date,
            lastDate: date.add(const Duration(days: 365)),
            fieldHintText: field.label ?? field.name,
            onDateSaved: (value) {
              args[field.name] = value.toIso8601String();
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.switchKind:
          children.add(Row(children: [
            Switch(
              value: false,
              onChanged: (value) {
                args[field.name] = value;
                widget.onChanged(args);
              },
            ),
            Text(field.label ?? field.name)
          ]));
          break;
        case FieldKind.check:
          children.add(Row(children: [
            Checkbox(
              value: false,
              onChanged: (value) {
                args[field.name] = value;
                widget.onChanged(args);
              },
            ),
            Text(field.label ?? field.name)
          ]));
          break;
        case FieldKind.text:
          children.add(TextFormField(
            keyboardType: TextInputType.multiline,
            decoration: InputDecoration(hintText: field.label ?? field.name),
            autofocus: field == widget.fields.first,
            onChanged: (value) {
              args[field.name] = value;
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.dateTime:
          children.add(DateTimeFormField(
            decoration: InputDecoration(
              suffixIcon: const Icon(Icons.event_note),
              labelText: field.label ?? field.name,
            ),
            mode: DateTimeFieldPickerMode.date,
            autovalidateMode: AutovalidateMode.always,
            onDateSelected: (DateTime value) {
              args[field.name] = value.toIso8601String();
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.password:
          children.add(TextFormField(
            keyboardType: TextInputType.visiblePassword,
            obscureText: true,
            decoration: InputDecoration(hintText: field.label ?? field.name),
            autofocus: field == widget.fields.first,
            onChanged: (value) {
              args[field.name] = num.parse(value);
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.address:
          children.add(AddressField(
            onChanged: (address, placeId) {
              args[field.name] = address;
              widget.onChanged(args);
            },
          ));
          break;
        case FieldKind.number:
        case FieldKind.percent:
        case FieldKind.currency:
          children.add(TextFormField(
            keyboardType: TextInputType.number,
            inputFormatters: [FilteringTextInputFormatter.digitsOnly],
            decoration: InputDecoration(hintText: field.label ?? field.name),
            autofocus: field == widget.fields.first,
            onChanged: (value) {
              args[field.name] = num.parse(value);
              widget.onChanged(args);
            },
          ));
          break;

        default:
          TextInputType keyboardType;
          switch (field.kind) {
            case FieldKind.phoneNumber:
              keyboardType = TextInputType.phone;
              break;
            case FieldKind.emailAddress:
              keyboardType = TextInputType.emailAddress;
              break;
            case FieldKind.url:
              keyboardType = TextInputType.url;
              break;
            default:
              keyboardType = TextInputType.text;
          }
          children.add(TextFormField(
            keyboardType: keyboardType,
            decoration: InputDecoration(hintText: field.label ?? field.name),
            autofocus: field == widget.fields.first,
            onChanged: (value) {
              args[field.name] = value;
              widget.onChanged(args);
            },
          ));
      }
    }
    return Column(
      children: children,
    );
  }
}
